#!/usr/bin/perl
# $crtd:  by  Derald Metzger  on  200223 $
# $cmnt:  Cfm.pm module. This module is part of the cfm pkg.
#    Copyright (C) 1998 - 2000 Derald Metzger
#    This may be freely redistributed under the terms of the GNU GPL.
# $

package Cfm;

use strict;

#  Load perl modules
use Cwd;
use File::Basename;
use File::Path;
use File::Spec;

#  Declare global vars
use vars '$archs';        # Supported archs
use vars '$cfmroot';      # Root of the cfmdb dir tree
use vars '$cfm_ver';      # Cfm-version
use vars '$cfm_ver_rel';  # Cfm-version-release
use vars '$dcs';          # Supported dataclasses
use vars '@def_sets';     # Default sets
use vars '$docdir';       # Cfm doc dir
use vars '$hnam';         # The executing host's name
use vars '$os_dirs';      # Supported os proper dir trees
use vars '$osv';          # Os and major rev
use vars '$osv_dc';       # Supported archs and dataclasses
use vars '$sdate';        # Short form date, yymmdd
use vars '@uname';        # Executing host's uname -a output
use vars '$vb_lvl';       # Verbosity level

#  Take care of env stuff
$cfmroot = $ENV{CFMROOT} || die "\n### \$CFMROOT not set. Exit";
@uname = split(/ /, `uname -a`);   # Get full uname string
$hnam = $uname[1];                 # Executing host's name

#  Id/set vars
my @args;                                  # System call args
$cfm_ver_rel = "cfm-1.0.0-1";         # Build fills this in
($cfm_ver) = $cfm_ver_rel =~ /(.*)-\d+/;   # Init to just name-ver
my $configfn = "$cfmroot/CFMROOT/config";  # Init config file
$docdir = "/usr/doc/$cfm_ver";             # Should be filled in by build
(my $osmaj, $_) = split /\./, $uname[2];   # Os major rev
$osv = "$uname[0]-$osmaj";                 # Os & major rev
my $rc;                                    # Return code for system calls
chop($sdate = `date +%y%m%d`);             # Short date yymmdd

#  Get these params from the Cfm main config file
if(-f $configfn) {
    eval `grep '^\$archs =' "$configfn"`;      # Get os proper dir trees
    eval `grep '^\$dcs =' "$configfn"`;        # Get os proper dir trees
    eval `grep '^\@def_sets =' "$configfn"`;   # Get default sets list
    eval `grep '^\$os_dirs =' "$configfn"`;    # Get os proper dir trees
    eval `grep '^\$osv_dc =' "$configfn"`;     # Get archs & dataclasses
    eval `grep '^\$vb_lvl =' "$configfn"`;     # Get verbosity level fm config
} else {
    print "## WARNING: config file not available. Normal at --initdb.\n";
}

#####################  subroutines and formats  #####################

#---------- vt_nodes($ia_r, $sfx, $err_r, $oa_r)
#           validate and type filesystem nodes
#Input:
#   This routine reads a list of filesystem nodes from an array,
# usually @ARGV. It removes, canonifies the path, validates, and
# determines their type.  Processing terminates when the input
# encounters an entry that begins with a `-' or a newline. Note that
# entries returned by this routine are legal for creation in the
# cfmdb.
#Validation:
#   An entry must reside in a valid dir tree to be valid.  Valid dir
# trees come in 2 flavors: os proper and other.  Os proper dirs are
# validated by verifying their top dir is in $os_dirs.  If in an
# $os_dirs dir tree a prefix (os-major_version) is inserted in the
# output to be used in locating the dir tree inside the cfmdb.
# Validating other dir trees proceeds by first stripping a possible
# mount point (/mnt<partition>). A mapped user access path not
# containing the partition mount point is used for all access.  The
# top dir of the mapped path may optionally contain an architecture id
# dir which must be in the $archs list. After testing, then then
# temporarily ignoring dirs that meet this requirement the next dir
# must be a dataclass in the $dcs list. All this supports filesystem
# resource management and servers in heterogeneous environments. Tho
# discouraged, you may bypass the arch and dataclass tests which
# supports heterogeneous server environments by putting all top level
# dirs in $os_dirs. See doc/assumptions.
# Types:
#   Once the node is validated its type is determined and the output
# suffix is provided to id the filenode in the cfmdb. If the caller
# provided a null suffix arg the node must exist on the host and its
# type is determined by testing. Otherwise these are legal:
#   input $sfx = ,v        regular file, output $sfx is ,v
#   input $sfx = .cfmsl,v  symlink,      output $sfx is .cfmsl,v
#   input $sfx = d         dir,          output $sfx is null
# Output:
#   Output entries for each node are 3 element lists of the form:
#  <arch prefix for os_dirs> <mapped_access_path> <output suffix>
#  @i_aref refs the array of input args
#  @o_aref refs the array of output args
#
# A bad arg increments the err count.
#
sub vt_nodes($ $ $ $) {
    my ($ia_r, $sfx, $err_r, $oa_r) = @_;
    my $known = $sfx;  # Null means test for type below
    my $fn;            # Local host access path
    ($sfx eq "d") && ($sfx = ());  # Null $sfx out for dirs

    while(($_ = $$ia_r[0]) && !/^-/) {
        my $pfx = ();      # Assume not os
        $fn = shift @$ia_r;
        ($fn eq '.') && ($fn = ());
        $fn =~ m;^/;o || ($fn = Cwd::cwd() . "/$fn");  # Make relative absolute
        ($fn) = File::Spec->canonpath($fn);    # Strip //, trailing /
        $fn =~ s;([^/]+/\.\.)|(/\.)($|/);;og;  # Strip `dir/..', `/.'
        if($fn =~ m;${os_dirs};o) { $pfx = "/$osv"; }
        else {  # Not in os dirs
            $fn =~ m;^(/mnt[^/]+)(/.+);o && ($fn = $2); # Strip mountpoint
            my ($arch, $dc) = $fn =~ m;${archs}(.*);o;  # Arch & dataclass
            (!$arch || ($arch && $dc)) && do {# Test dataclass, lone arch ok
                $dc = $arch ? "/$dc" : $fn;  # ...
                $dc =~ m#${dcs}#o || do {  # Not in a dataclass
                    warn "### Reject unsupported dirtree: $fn\n",
				    "    NOT in \$os_dirs or [\$archs]\$dcs lists\n";
				    $$err_r++; next;
                };
            };
        }
        $known || do {  # determine type by testing
            $sfx = ();
            if(-l $fn) { $sfx = ".cfmsl,v"; }  # symlinks get .cfmsl,v
            elsif(-f $fn) { $sfx = ",v"; }     # regular file gets ,v
            elsif(-d $fn) { ; }  # Its a dir, pass it thru
            else {
                warn "### Reject! Not symlink,reg,dir: $fn\n";
				$$err_r++; next;
            }
        };
        push @$oa_r, [ $pfx, $fn, $sfx ];
    }
    return 1;
}

#----------
# This routine returns the name of the 1st set in the host's set list
# that has a branch in the rcsfile along with rev info for that set in
# the form of <branch> and <level>. Cfm uses one branch for each set
# with the set name tagging the head of its branch, eg after init you
# have `<set>: 1.1.<x>.1' where <x> determines the branch and the final
# 1 is the level.
# Input: 
#   $dbf  = cfmdb rcsfile to check
#   @sets = List of sets to check for
# Output: success
#   return    = version level of set found
#   $$br_ref  = version branch of set found
#   $$set_ref = 1st set name found
# Output: No set found
#   return    = 0
#   $$br_ref  = Next available branch
#   $$set_ref = null
# Output: Failed
#   return    = 0
#   $$br_ref  = null
#   $$set_ref = null

sub chk_rev($ $ $ @) {
    my ($dbf, $set_ref, $br_ref, @sets) = @_;
    my $rsym_flg = 0;    # Found set (symbol) lines
    my @rsymlines = ();  # Captured set lines
    $$set_ref = "";      # Init to null
    $$br_ref = 0;       # Init to 0

    # Get symbol lines. These hold the set names & rcs revisions
    open(RCSFILE, "$dbf") || do {
        warn "\### Failed: open $dbf\n";
        return 0;
    };
    while(<RCSFILE>) {
        $rsym_flg || /^symbols[;]?$/ || next;
        $rsym_flg || /^symbols[;]?$/ && ($rsym_flg = 1) && next;
        /^locks[;]?/ && last;
        push @rsymlines, $_;
    }
    close(RCSFILE);
    
    # Find 1st matching set name. Get its set, branch, & version
    for my $set (@sets) {
        my $set1 = $set;
        $set1 =~ s/\+/\\+/;
        for my $symline (@rsymlines) {
            $symline =~ m/\t$set1:(1\.1\.\d+)\.(\d+).*/
                && do {$$set_ref=$set;$$br_ref=$1;return $2;};
        }
    }
    my $br;
    foreach my $symline (@rsymlines) { # Get bigest branch in use
        $symline =~ m/[^:]+:1\.1\.(\d+).*/ && ($1 > $br) && ($br = $1);
    }
    $$br_ref = "1.1." . ++$br;
    return 0;
}

#-------------
# This routine erases a set membership from a cfmdb rcsfile.

sub rm_setmem($ $) {
    my ($dbf, $set) = @_;

    @args = ("rcs", "-q", "-o:$set", "$dbf");  # Remove branch
    ($rc = system(@args)) && do {
        ($vb_lvl > 3) && bad_syscall();
        return 0;
    };
    @args = ("rcs",  "-q", "-N$set", "$dbf");
    ($rc = system(@args)) && do {  # Remove set symbol
        ($vb_lvl > 3) && bad_syscall();
        return 0;
    };
    return 1;
}

#----------
# This routine will print info about a failed system call.
# It assumes the call has just executed: ($rc = system(@args)).
# Its handy for debugging. Note $rc != 0 for failures.

sub bad_syscall() {
    warn "### Failed: system(@args)";
    printf "\n    returned: %#04x\n", $rc;
}
